#!/usr/bin/env python3
"""
Small library for writing a collection of protobuf objects in a binary file

File format:
The file starts with a magic number
Then a tuple of type, length and value are written to file
- type encoded as int8
- size encoded as varint32

After the magic number, the file descriptor of the protobufs is written.
The messages passed to the function are then written to file, with first
their descriptor full name and then their value.
"""

from pbzlib.reader import PBZReader
from pbzlib.writer import PBZWriter


def write_pbz(fname, fdescr, *msgs):
    w = PBZWriter(fname, fdescr)
    if len(msgs) == 0:
        # Returns writer to caller
        return w
    else:
        # Directly write the messages to file and close
        for msg in msgs:
            w.write(msg)
        w.close()


def open_pbz(fname):
    sentinel = object()
    r = PBZReader(fname)
    while True:
        obj = r.next(sentinel)
        if obj is sentinel:
            return
        yield obj



import gzip
import google.protobuf
from google.protobuf import descriptor_pool
from google.protobuf.descriptor_pb2 import FileDescriptorSet
from google.protobuf.internal.encoder import _VarintEncoder

from pbzlib.reader import PBZReader
from pbzlib.constants import MAGIC, T_PROTOBUF_VERSION, T_FILE_DESCRIPTOR, T_DESCRIPTOR_NAME, T_MESSAGE


class PBZWriter:
    def __init__(self, fname, fdescr, compresslevel=9, autoflush=False):
        self._ve = _VarintEncoder()
        self._last_descriptor = None
        self.autoflush = autoflush

        self._fobj = gzip.open(fname, "wb", compresslevel=compresslevel)
        self._write_header(fdescr)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def _write_header(self, fdescr):
        self._fobj.write(MAGIC)

        # Write protocol buffer version in header
        self._write_blob(T_PROTOBUF_VERSION, len(google.protobuf.__version__), google.protobuf.__version__.encode("utf8"))

        if type(fdescr) == PBZReader:
            self._write_blob(T_FILE_DESCRIPTOR, len(fdescr._raw_descriptor), (fdescr._raw_descriptor))
            self._dpool = fdescr._dpool

        else:
            # Read FileDescriptorSet
            with open(fdescr, "rb") as fi:
                fdset = fi.read()
                sz = fi.tell()

            # Write FileDescriptorSet
            self._write_blob(T_FILE_DESCRIPTOR, sz, fdset)

            # Parse descriptor for checking that the messages will be defined in
            # the serialized file
            self._dpool = descriptor_pool.DescriptorPool()
            ds = FileDescriptorSet()
            ds.ParseFromString(fdset)
            for df in ds.file:
                self._dpool.Add(df)

    def _write_blob(self, mtype, size, value):
        self._fobj.write(bytes([mtype]))
        self._ve(self._fobj.write, size)
        self._fobj.write(value)
        if self.autoflush:
            self._fobj.flush()

    def close(self):
        """
        Close PBZ file
        """
        self._fobj.close()

    def write(self, msg):
        """
        Writes a protobuf message to the file
        """
        if msg.DESCRIPTOR.full_name != self._last_descriptor:
            # Check that the message type has been defined
            self._dpool.FindMessageTypeByName(msg.DESCRIPTOR.full_name)

            self._write_blob(T_DESCRIPTOR_NAME, len(msg.DESCRIPTOR.full_name), msg.DESCRIPTOR.full_name.encode("utf8"))
            self._last_descriptor = msg.DESCRIPTOR.full_name

        self._write_blob(T_MESSAGE, msg.ByteSize(), msg.SerializeToString())



#!/usr/bin/env python3

import gzip
import warnings
import google.protobuf
from google.protobuf import descriptor_pool
from google.protobuf import reflection
from google.protobuf.internal.decoder import _DecodeVarint
from google.protobuf.descriptor_pb2 import FileDescriptorSet

from pbzlib.constants import MAGIC, T_PROTOBUF_VERSION, T_FILE_DESCRIPTOR, T_DESCRIPTOR_NAME, T_MESSAGE


class PBZReader:
    def __init__(self, fname, return_raw_object=False):
        self.return_raw_object = return_raw_object

        self._next_descr = None
        self._next_descr_name = None

        self._prevbuf = b""
        self._fobj = gzip.open(fname, "rb")
        assert self._fobj.read(len(MAGIC)) == MAGIC
        self._dpool, self._raw_descriptor = self.read_descriptor_pool()

    def _read_next_obj(self):
        """
        Each object is encoded as:
         - vtype (uint8_t): type of the object
         - size (varint): size of the object as varint
         - object itself
        """
        bufsz = self._prevbuf + self._fobj.read(9 - len(self._prevbuf))
        if len(bufsz) == 0:
            return None, None

        vtype = bufsz[0]
        size, pos = _DecodeVarint(bufsz[1:], 0)

        if pos + size < 8:
            data = bufsz[1 + pos: 1 + pos + size]
            self._prevbuf = bufsz[1 + pos + size:]
        elif pos + size < 8:
            data = bufsz[1:]
            self._prevbuf = b""
        else:
            data = bufsz[1+pos:] + self._fobj.read(size - 8 + pos)
            self._prevbuf = b""

        return vtype, data

    def read_descriptor_pool(self):
        dpool = descriptor_pool.DescriptorPool()
        while True:
            vtype, data = self._read_next_obj()
            if vtype is None:
                raise Exception("Unexpected end of file")

            if vtype == T_FILE_DESCRIPTOR:
                ds = FileDescriptorSet()
                ds.ParseFromString(data)
                for df in ds.file:
                    dpool.Add(df)
                return dpool, data

            elif vtype == T_PROTOBUF_VERSION:
                pbversion = data.decode("utf8")
                if google.protobuf.__version__.split(".") < pbversion.split("."):
                    warnings.warn(f"File uses more recent of protobuf ({pbversion})")

            else:
                raise Exception(f"Unknown message type {vtype}")

    def next(self, default=None):
        while True:
            vtype, data = self._read_next_obj()
            if vtype is None:
                if default is None:
                    raise StopIteration
                return default

            if vtype == T_DESCRIPTOR_NAME:
                self._next_descr_name = data.decode("utf8")
                self._next_descr = self._dpool.FindMessageTypeByName(self._next_descr_name)

            elif vtype == T_MESSAGE:
                if self.return_raw_object:
                    return self._next_descr_name, self._next_descr, data
                else:
                    return reflection.ParseMessage(self._next_descr, data)

            else:
                raise Exception(f"Unknown message type {vtype}")



#!/usr/bin/env python

from setuptools import setup, find_packages
from codecs import open

with open('README.md', 'r', 'utf-8') as fd:
    long_description = fd.read()

setup(
    name='pbzlib',
    version='0.20210128',
    description='Library for serializing a list of protobuf objects',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/fabgeyer/pbzlib',
    author='fabgeyer',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
    ],
    keywords='protobuf',
    packages=find_packages(),
    install_requires=['protobuf'],
)
